from PySide6.QtWidgets import QListWidget
from PySide6.QtWidgets import QLabel
from PySide6 import QtCore
import traceback

class ListBox(QListWidget):

    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.element = element
        resize_factor = element.gui.resize_factor

        self.add_items()

        # Adjust list value types
        if element.select == "item":
            if not self.element.option_value.variable:
                # List
                for i, val in enumerate(element.option_value.list):
                    if element.type == float:
                        element.option_value[i] = float(val)
                    elif element.type == int:
                        element.option_value[i] = int(val)

        if element.text:
            if type(element.text) == str:
                txt = element.text
            else:
                txt = self.element.getvar(element.text.variable_group, element.text.variable)    
            label = QLabel(txt, element.parent.widget)
            label.setStyleSheet("background: transparent; border: none")
            if not element.enable:
                label.setEnabled(False)
            self.text_widget = label
            label.setVisible(True)

        if self.element.tooltip:
            if type(self.element.tooltip) == str:
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)
        
        if self.element.multiselection:
            self.setSelectionMode(3)

        # First call back to change the variable
        self.itemSelectionChanged.connect(self.callback)
#        self.itemClicked.connect(self.callback)

        self.set_geometry()

        self.execute_callback = True

    def set(self):

        self.execute_callback = False

        # First check if items need to be updated. This is only necessary when "option_string" is a dict
        if self.element.option_string.variable:
            self.add_items()

        # Get the items
        items = []
        for x in range(self.count()):
            items.append(self.item(x))

        # Get value(s). If self.element.variable is not defined (i.e. None), we can skip setting the index
        if self.element.variable is not None:
            
            # Get the variable
            val = self.element.getvar(self.element.variable_group, self.element.variable)

            # Now get the options
            if self.element.select == "item":
                if self.element.option_value.variable:
                    name  = self.element.option_value.variable
                    group = self.element.option_value.variable_group
                    vals = self.element.getvar(group, name)
                    if not vals:
                        vals = [""]
                else:
                    vals = self.element.option_value.list


            if self.element.multiselection:
                for index, item in enumerate(items):
                    if self.element.select == "item":
                        if vals[index] in val:   
                            item.setSelected(True)
                        else:
                            item.setSelected(False)
                    else:
                        if index in val:   
                            item.setSelected(True)
                        else:
                            item.setSelected(False)

            else:    

                # Now get the values
                if self.element.select == "item":
                    if val in vals:
                        index = vals.index(val)
                    else:
                        index = 0
                        if type(val) is not str:
                            valstr = str(val)
                        else:
                            valstr = val
                        print(f"Error in listbox: variable {self.element.variable} ({valstr}) not found in list!")
                else:
                    index = val
                self.setCurrentItem(items[index])

            self.execute_callback = True    

        if type(self.element.text) != str:
            txt = self.element.getvar(self.element.text.variable_group, self.element.text.variable)
            self.text_widget.setText(txt)

        if type(self.element.tooltip) != str:
            txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

    def callback(self):

        if not self.execute_callback:
            return

        if self.element.multiselection:

            newval  = []

            if self.element.select == "item":
                if self.element.option_value.variable:
                    name = self.element.option_value.variable
                    group = self.element.option_value.variable_group
                    vals = self.element.getvar(group, name)
                    if not vals:
                        vals = [""]
                else:
                    vals = self.element.option_value.list

            for indx in self.selectedIndexes():
                if self.element.select == "item":
                    newval.append(vals[indx.row()])
                else:
                    newval.append(indx.row())    

        else:    

            index = self.currentRow()
            if self.element.select == "item":
                if self.element.option_value.variable:
                    name = self.element.option_value.variable
                    group = self.element.option_value.variable_group
                    vals = self.element.getvar(group, name)
                    if not vals:
                        vals = [""]
                else:
                    vals = self.element.option_value.list
                newval = vals[index]
            else:
                newval = index

        name  = self.element.variable
        group = self.element.variable_group
        self.element.setvar(group, name, newval)

        try:
            if self.isEnabled() and self.element.callback:
                self.element.callback(newval, self)
            # Update GUI
            self.element.window.update()
        except:
            traceback.print_exc()

    def add_items(self):

        # Delete existing items
        self.clear()

        if self.element.option_string.variable:
            group = self.element.option_string.variable_group
            name  = self.element.option_string.variable
            v     = self.element.getvar(group, name)
            if not v:
                v = [""]
            for itxt, txt in enumerate(v):
                self.insertItem(itxt, txt)
        else:
            for itxt, txt in enumerate(self.element.option_string.list):
                self.insertItem(itxt, txt)

    def set_geometry(self):

        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)
        if self.element.text:
            label = self.text_widget
            fm = label.fontMetrics()
            wlab = int(fm.size(0, self.element.text).width())
            if self.element.text_position == "above-center" or self.element.text_position == "above":
                label.setAlignment(QtCore.Qt.AlignCenter)
                label.setGeometry(x0, int(y0 - 20 * resize_factor), wdt, int(20 * resize_factor))
            elif self.element.text_position == "above-left":
                label.setAlignment(QtCore.Qt.AlignLeft)
                label.setGeometry(x0, int(y0 - 20 * resize_factor), wlab, int(20 * resize_factor))
            else:
                # Assuming left
                label.setAlignment(QtCore.Qt.AlignRight)
                label.setGeometry(int(x0 - wlab - 3 * resize_factor),
                                  int(y0 + 5 * self.element.gui.resize_factor),
                                  wlab,
                                  int(20 * resize_factor))



import os
import time

import rasterio
import rasterio.features
from rasterio.warp import calculate_default_transform, reproject, Resampling, transform_bounds
from rasterio import MemoryFile
from rasterio.transform import Affine
import shutil

from .layer import Layer

class ImageLayer(Layer):
    def __init__(self, map, id, map_id):
        super().__init__(map, id, map_id)
        self.active = False
        self.type   = "image"
        self.file_name = map_id + ".png"

    def activate(self):
        self.active = True
        self.show()

    def deactivate(self):
        self.active = False

    def clear(self):
        self.active = False
        # js_string = "import('/js/main.js').then(module => {module.removeLayer('" + self.map_id + "')});"
        # self.map.view.page().runJavaScript(js_string)
        self.map.runjs("/js/main.js", "removeLayer", arglist=[self.map_id])


    def make_overlay(self):    
        fname = os.path.join(self.map.server_path, "overlays", self.file_name)
        coords = self.map.map_extent
        xlim = [coords[0][0], coords[1][0]]
        ylim = [coords[0][1], coords[1][1]]
        width = self.map.view.geometry().width()
        okay = self.data.map_overlay(fname, xlim=xlim, ylim=ylim, width=width)
        if okay:
            return xlim, ylim
        else:
            return None, None

    def update(self):
        if hasattr(self.data, "map_overlay"):
            xlim, ylim = self.make_overlay()
            if xlim is None:
                return
            # if xlim[0] > xlim[1]:
            #     xlim[0] -= 360.0
            bounds = [[xlim[0], xlim[1]], [ylim[0], ylim[1]]]
            overlay_file = f"./overlays/{self.file_name}"
            self.map.runjs("/js/image_layer.js", "updateLayer", arglist=[overlay_file, self.map_id, bounds])
            self.map.runjs("/js/image_layer.js", "setOpacity", arglist=[self.map_id, 1.0])

    def set_data(self, data, image_file=None, xlim=None, ylim=None):

        self.data = data
        self.map.runjs("/js/main.js", "removeLayer", arglist=[self.map_id])

        try:
            xlim, ylim = self.make_overlay()
            if xlim is None:
                return
        except Exception as e:
            print(f"Something went wrong with map overlay: {self.map_id}, {e}")
            return

        bounds = [[xlim[0], xlim[1]], [ylim[0], ylim[1]]]
        overlay_file = f"./overlays/{self.file_name}"
        self.map.runjs("/js/image_layer.js", "addLayer", arglist=[self.map_id])
        self.map.runjs("/js/image_layer.js", "updateLayer", arglist=[overlay_file, self.map_id, bounds])

    # def set_data(self,
    #              data,
    #              image_file=None,
    #              xlim=None,
    #              ylim=None):
        
    #     fname = os.path.join(self.map.server_path, "overlays", self.file_name)

    #     self.data = data

        # id_string = self.id

        # dataset = rasterio.open(image_file)

        # src_crs = 'EPSG:4326'
        # dst_crs = 'EPSG:3857'

        # with rasterio.open(image_file) as src:
        #     transform, width, height = calculate_default_transform(
        #         src.crs, dst_crs, src.width, src.height, *src.bounds)
        #     kwargs = src.meta.copy()
        #     kwargs.update({
        #         'crs': dst_crs,
        #         'transform': transform,
        #         'width': width,
        #         'height': height
        #     })
        #     bnds = src.bounds

        #     mem_file = MemoryFile()
        #     with mem_file.open(**kwargs) as dst:
        #         for i in range(1, src.count + 1):
        #             reproject(
        #                 source=rasterio.band(src, i),
        #                 destination=rasterio.band(dst, i),
        #                 src_transform=src.transform,
        #                 src_crs=src.crs,
        #                 dst_transform=transform,
        #                 dst_crs=dst_crs,
        #                 resampling=Resampling.nearest)

        #         # band1 = dst.read(1)

        # new_bounds = transform_bounds(dst_crs, src_crs,
        #                               dst.bounds[0],
        #                               dst.bounds[1],
        #                               dst.bounds[2],
        #                               dst.bounds[3])
        # isn = np.where(band1 < 0.001)
        # band1[isn] = np.nan

        # band1 = np.flipud(band1)
        # cminimum = np.nanmin(band1)
        # cmaximum = np.nanmax(band1)

        # norm = matplotlib.colors.Normalize(vmin=cminimum, vmax=cmaximum)
        # vnorm = norm(band1)

        # cmap = cm.get_cmap(colormap)
        # im = Image.fromarray(np.uint8(cmap(vnorm) * 255))

        # shutil.copy(image_file, os.path.join(self.map.server_path, "overlays", self.file_name))

        # js_string = "import('/js/main.js').then(module => {module.removeLayer('" + self.map_id + "')});"
        # self.map.view.page().runJavaScript(js_string)

        # try:
        #     xlim, ylim = self.make_overlay()
        #     if xlim is None:
        #         return
        # except:
        #     print("Something went wrong with map overlay : " + self.map_id)
        #     return
        # bounds = [[xlim[0], xlim[1]], [ylim[0], ylim[1]]]
        # bounds_string = "[[" + str(bounds[0][0]) + "," + str(bounds[0][1]) + "],[" + str(bounds[1][0]) + "," + str(bounds[1][1]) + "]]"
        # overlay_file = "./overlays/" + self.file_name
        # js_string = "import('/js/image_layer.js').then(module => {module.addLayer('" + overlay_file + "','" + self.map_id + "'," + bounds_string + ","")});"
        # self.map.view.page().runJavaScript(js_string)


#         overlay_file = "./overlays/" + "main.background_topography.png"

#         # Bounds
#         bounds = [[xlim[0], xlim[1]], [ylim[0], ylim[1]]]
#         bounds_string = "[[" + str(bounds[0][0]) + "," + str(bounds[0][1]) + "],[" + str(bounds[1][0]) + "," + str(bounds[1][1]) + "]]"

# #        overlay_file = ""

#         js_string = "import('/js/image_layer.js').then(module => {module.addLayer('" + overlay_file + "','" + self.map_id + "'," + bounds_string + ","")});"
#         self.map.view.page().runJavaScript(js_string)

# #        self.update()




from .layer import Layer
from geopandas import GeoDataFrame

class HeatmapLayer(Layer):
    def __init__(self, map, id, map_id, **kwargs):
        super().__init__(map, id, map_id, **kwargs)
        pass

    def set_data(self, data, density_property=""):
        self.data = data
        self.density_property = density_property
        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                self.data = GeoDataFrame()
        self.visible = True 
        # Add new layer
        self.map.runjs(
            "/js/heatmap_layer.js",
            "addLayer",
            arglist=[
                self.map_id,
                self.data,
                self.max_zoom,
                self.density_property,
                self.side
            ],
        )

    def redraw(self):
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data)
        if not self.get_visibility():
            self.set_visibility(False)

    # def activate(self):
    #     self.active = True

    # def deactivate(self):
    #     self.active = False

    # def clear(self):
    #     self.active = False
    #     self.remove()

    # def remove(self):
    #     self.map.runjs(self.main_js, "removeLayer", arglist=[self.map_id, self.side])

    # def update(self):
    #     # This layer does not get updated when zooming in or out
    #     pass

    # def set_visibility(self, true_or_false):
    #     if true_or_false:
    #         self.map.runjs("/js/main.js", "showLayer", arglist=[self.map_id])
    #     else:
    #         self.map.runjs("/js/main.js", "hideLayer", arglist=[self.map_id])



#import os
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cm
import json

class MplColorHelper:

  def __init__(self, cmap_name, start_val, stop_val):
    self.cmap_name = cmap_name
    self.cmap = plt.get_cmap(cmap_name)
    self.norm = mpl.colors.Normalize(vmin=start_val, vmax=stop_val)
    self.scalarMap = cm.ScalarMappable(norm=self.norm, cmap=self.cmap)

  def get_rgb(self, val):
      rgb0   = self.scalarMap.to_rgba(val)
      rgb    = [0, 0, 0]
      rgb[0] = round(rgb0[0] * 255)
      rgb[1] = round(rgb0[1] * 255)
      rgb[2] = round(rgb0[2] * 255)
      return rgb

class ColorBar:
    def __init__(self, colormap="jet", scale="linear", orientation="vertical", legend_title="", color_values=None):
        self.title       = legend_title
        self.orientation = orientation
        self.color_map   = colormap
        self.scale       = scale
        self.contour     = []
        self.color_values = color_values
            

    def make(self,
             cmin,
             cmax,
             cstep=None,
             decimals=None,
             reverse=True):
        
        if self.color_values:
            for val in self.color_values:

                zl = val["lower_value"]
                zu = val["upper_value"]

                if decimals:
                    zu = np.around(zu, decimals)
                    zl = np.around(zl, decimals)

                zustr = str(zu)
                zlstr = str(zl)

                if decimals==0:
                    zustr = str(int(zu))
                    zlstr = str(int(zl))
                elif decimals is not None:
                    zustr = str(np.around(zu, decimals))
                    zlstr = str(np.around(zl, decimals))

                contour = {}
                if "string" in val:
                    contour["string"] = val["string"]
                else:    
                    if zu>1.0e7:
                        contour["string"] = "> " + zlstr
                    elif zl<-1.0e7:
                        contour["string"] = "< " + zustr
                    else:    
                        contour["string"] = zlstr + " - " + zustr

                contour["lower_value"] = zl
                contour["upper_value"] = zu

                contour["rgb"] = val["rgb"]
                contour["hex"] = rgb2hex(tuple(val["rgb"]))

                self.contour.append(contour)

        else:

            if np.isnan(cmin) or np.isnan(cmax):
                return

            clmap = MplColorHelper(self.color_map, cmin, cmax)

            if self.scale == "linear":
                if not cstep:
                    cstep = (cmax - cmin) / 20

                nsteps = round((cmax - cmin) / cstep + 2)

                for i in range(nsteps):

                    # Interpolate
                    zl = cmin + (i - 1) * cstep
                    zu = zl + cstep
                    z = 0.5 * (zl + zu)
                    rgb = clmap.get_rgb(z)

                    if decimals:
                        zu = np.around(zu, decimals)
                        zl = np.around(zl, decimals)

                    zustr = str(zu)
                    zlstr = str(zl)

                    if decimals==0:
                        zustr = str(int(zu))
                        zlstr = str(int(zl))
                    elif decimals is not None:
                        zustr = str(np.around(zu, decimals))
                        zlstr = str(np.around(zl, decimals))

                    contour = {}
                    if i == 0:
    #                    contour["string"] = "< " + str(cmin)
                        contour["string"] = "< " + zustr
                        contour["lower_value"] = -1.0e6
                        contour["upper_value"] = zu
                    elif i == nsteps - 1:
    #                    contour["string"] = "> " + str(cmax)
                        contour["string"] = "> " + zlstr
                        contour["lower_value"] = zl
                        contour["upper_value"] = 1.0e6
                    else:
                        contour["string"] = zlstr + " - " + zustr
                        contour["lower_value"] = zl
                        contour["upper_value"] = zu
                    contour["rgb"] = rgb
                    contour["hex"] = rgb2hex(tuple(rgb))

                    self.contour.append(contour)

            if reverse:
                self.contour.reverse()


    def to_json(self):
        jsn = {}
        jsn["title"] = self.title
        jsn["contour"] = []
        for cnt in self.contour:
            contour= {}
            contour["text"] = cnt["string"]
            contour["color"] = "#" + cnt["hex"]
            jsn["contour"].append(contour)
        return json.dumps(jsn)

    def to_dict(self):
        jsn = {}
        jsn["title"] = self.title
        jsn["contour"] = []
        for cnt in self.contour:
            contour= {}
            contour["text"] = cnt["string"]
            contour["color"] = "#" + cnt["hex"]
            jsn["contour"].append(contour)
        return jsn

def rgb2hex(rgb):
    return '%02x%02x%02x' % rgb




from PySide6.QtWidgets import QLineEdit
from PySide6.QtWidgets import QLabel
from PySide6 import QtCore
from PySide6.QtCore import Qt
import traceback

class Text(QLabel):

    def __init__(self, element):
        super().__init__("", element.parent.widget)

        self.element = element

        self.setVisible(True)

        self.set_geometry()
        
        if self.element.text_position:
            self.setAlignment(Text.convert_to_qt_alignment(self.element.text_position))

    def set(self):
        if self.element.variable:
            group = self.element.variable_group
            name = self.element.variable
            val = self.element.getvar(group, name)
        else:
            val = self.element.text
        self.setText(val)

    def set_geometry(self):
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)

    def convert_to_qt_alignment(alignment_string):
        alignment_mapping = {
            'left': Qt.AlignLeft | Qt.AlignVCenter,	# This is also the default for when nothing is specified
            'right': Qt.AlignRight | Qt.AlignVCenter,
            'top': Qt.AlignTop | Qt.AlignHCenter,
            'bottom': Qt.AlignBottom | Qt.AlignHCenter,
            'top-left': Qt.AlignTop | Qt.AlignLeft,
            'top-right': Qt.AlignTop | Qt.AlignRight,
            'bottom-left': Qt.AlignBottom | Qt.AlignLeft,
            'bottom-right': Qt.AlignBottom | Qt.AlignRight,
            'center': Qt.AlignCenter
        }

        return alignment_mapping.get(alignment_string, Qt.AlignLeft | Qt.AlignVCenter)


from PyQt5.QtWidgets import QSplashScreen
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QApplication
import os

class Splash:

    def __init__(self, splash_file, seconds=20.0):
        # Splash screen
        QSplashScreen(QPixmap(splash_file))
        self.splash = QSplashScreen(QPixmap(splash_file), Qt.WindowStaysOnTopHint)
        self.splash.show()
        # Close SplashScreen after 20 seconds (20,000 ms)
        QTimer.singleShot(int(seconds*1000), self.splash.close)

    def close(self):
        self.splash.close()



import os
from PyQt5.QtWidgets import QPushButton, QFileDialog
import traceback


class PushOpenDir(QPushButton):

    def __init__(self, element):
        super().__init__("", element.parent.widget)

        self.element = element

        if element.text:
            if isinstance(element.text, str):
                txt = element.text
            else:
                txt = self.element.getvar(
                    element.text.variable_group, element.text.variable
                )
            self.setText(txt)

        if self.element.tooltip:
            if isinstance(element.tooltip, str):
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(
                    self.element.tooltip.variable_group, self.element.tooltip.variable
                )
            self.setToolTip(txt)

        self.setVisible(True)

        self.clicked.connect(self.callback)

        self.set_geometry()

    def set(self):
        if not isinstance(self.element.text, str):
            self.setText(
                self.element.getvar(
                    self.element.text.variable_group, self.element.text.variable
                )
            )
        if not isinstance(self.element.tooltip, str):
            self.setToolTip(
                self.element.getvar(
                    self.element.tooltip.variable_group, self.element.tooltip.variable
                )
            )

    def callback(self):
        self.okay = True
        group = self.element.variable_group
        name = self.element.variable
        val = self.element.getvar(group, name)
        if not val:
            val = os.getcwd()
        val = str(val).replace("\\", "/")
        dir_name = QFileDialog.getExistingDirectory(
            parent=self,
            caption=self.element.title,
            directory=val,
            options=QFileDialog.ShowDirsOnly,
        )
        if dir_name:
            self.element.setvar(group, name, dir_name)
        else:
            self.okay = False

        try:
            if self.okay and self.element.callback:
                self.element.callback(dir_name, self)
                self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)



from PySide6.QtWidgets import QLineEdit
from PySide6.QtWidgets import QLabel
from PySide6 import QtCore
import traceback

class Edit(QLineEdit):

    def __init__(self, element):
        super().__init__("", element.parent.widget)

        self.element = element

        self.string = None

        if element.type == float or element.type == int:
            self.setAlignment(QtCore.Qt.AlignRight)
        if not element.enable:
            self.setEnabled(False)

        self.setVisible(True)

        if element.text:
            if isinstance(element.text, str):
                txt = element.text
            else:
                txt = self.element.getvar(element.text.variable_group, element.text.variable)    
            label = QLabel(txt, element.parent.widget)
            label.setStyleSheet("background: transparent; border: none")
            if not element.enable:
                label.setEnabled(False)
            self.text_widget = label
            label.setVisible(True)

        if self.element.tooltip:
            if isinstance(self.element.tooltip, str):
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

        self.editingFinished.connect(self.callback)

        self.set_geometry()

    def set(self):
        group  = self.element.variable_group
        name   = self.element.variable
        val    = self.element.getvar(group, name)
        self.string = str(val)
        self.setText(str(val))
        self.setStyleSheet("")

        if not isinstance(self.element.text, str):
            txt = self.element.getvar(self.element.text.variable_group, self.element.text.variable)
            self.text_widget.setText(txt)

        if not isinstance(self.element.tooltip, str):
            txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

    def callback(self):
        self.okay = True
        newtext = self.text()
        if self.element.type == int:
            try:
                newval = int(newtext)
            except:
                self.okay = False
        elif self.element.type == float:
            try:
                newval = float(newtext)
            except:
                self.okay = False
        else:
            newval = newtext

        # Do some range checking here ...

        # Update value in variable dict
        if self.okay:
            group = self.element.variable_group
            name = self.element.variable
            self.element.setvar(group, name, newval)
            self.setStyleSheet("")
        else:
            self.setStyleSheet("QLineEdit{background-color: red}")

        if self.string == newtext:
            # Nothing changed
            self.okay = False # Not going to run second callback

        try:
            if self.okay and self.isEnabled() and self.element.callback:
                group = self.element.variable_group
                name = self.element.variable
                val   = newval
                self.element.callback(val, self)
                # Update GUI
            self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)
        if self.element.text:
            if not isinstance(self.element.text, str):
                txt = self.element.getvar(self.element.text.variable_group, self.element.text.variable)
            else:
                txt = self.element.text    
            label = self.text_widget
            fm = label.fontMetrics()
            wlab = int(fm.size(0, txt).width())
            if self.element.text_position == "above-center" or self.element.text_position == "above":
                label.setAlignment(QtCore.Qt.AlignCenter)
                label.setGeometry(x0, int(y0 - 20 * resize_factor), wdt, int(20 * resize_factor))
            elif self.element.text_position == "above-left":
                label.setAlignment(QtCore.Qt.AlignLeft)
                label.setGeometry(x0, int(y0 - 20 * resize_factor), wlab, int(20 * resize_factor))
            elif self.element.text_position == "right":
                label.setAlignment(QtCore.Qt.AlignLeft)
                label.setGeometry(int(x0 + wdt + 3 * resize_factor),
                                  int(y0 + 5 * self.element.gui.resize_factor),
                                  wlab,
                                  int(20 * resize_factor))
            else:
                # Assuming left
                label.setAlignment(QtCore.Qt.AlignRight)
                label.setGeometry(int(x0 - wlab - 3 * resize_factor),
                                  int(y0 + 5 * self.element.gui.resize_factor),
                                  wlab,
                                  int(20 * resize_factor))



from PySide6.QtWidgets import QSlider, QLabel, QStyle, QStyleOptionSlider
from PySide6.QtCore import Qt, QRect, QPoint
from PySide6.QtGui import QPainter
from PySide6 import QtWidgets
import traceback


class Slider(QSlider):

    def __init__(self, element):
        super().__init__(Qt.Horizontal, parent=element.parent.widget)

        self.element = element

        # ToDO: Make minimum and maximum flexible
        minimum = element.minimum
        maximum = element.maximum
        interval = 10
        self.setMinimum(minimum)
        self.setMaximum(maximum)
        self.setTickInterval(interval)
        self.setSingleStep(1)
        self.setTickPosition(QSlider.TicksBelow)

        if element.text:
            label = QLabel(element.text, self.element.parent.widget)
            label.setStyleSheet("background: transparent; border: none")
            self.text_widget = label

        # set tick labels
        self.tick_label = []
        for ii, ticklabel in enumerate([minimum,maximum]):
            label = QLabel(str(ticklabel), self.parent)
            label.setStyleSheet("background: transparent; border: none")
            self.tick_widget.append(label)

        self.valueChanged.connect(self.first_callback)
        self.sliderReleased.connect(self.second_callback)

        self.set_geometry()

    def set(self):
        group = self.element.variable_group
        name = self.element.variable
        val = self.element.getvar(group, name)
        self.setValue(val)

    def first_callback(self):
        self.okay = True
        val = self.value()
        # Update value in variable dict
        group = self.element.variable_group
        name = self.element.variable
        self.element.setvar(group, name, val)
        self.setValue(val)
        if self.slide_callback:
            self.slide_callback(val, self)

    def second_callback(self):
        try:
            if self.isEnabled() and self.element.callback:
                group = self.element.variable_group
                name  = self.element.variable
                val   = self.element.getvar(group, name)
                self.element.callback(val, self)
                # Update GUI
                self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)

        if self.element.text:
            label = self.text_widget
            fm = label.fontMetrics()
            wlab = fm.size(0, element.text).width() + 15
            label.setAlignment(Qt.AlignRight)
            label.setGeometry(x0 - wlab - 3, y0 + 5, wlab, hgt)

        # set tick labels
        ticks = [x0, x0 + wdt]
        for ii, label in enumerate(self.tick_widget):
            fm = label.fontMetrics()
            hlab = fm.height()
            wlab = fm.width(label.text, label.fontInfo().pointSize())
            label.setAlignment(Qt.AlignHCenter)
            label.setAlignment(Qt.AlignTop)
            label.setGeometry(ticks[ii], y0 + hgt, wlab, hlab)

# class LabeledSlider(WidgetGroup):
#     def __init__(self, element, parent):
#         super().__init__(element, parent)
#         # super(LabeledSlider, self).__init__(element, parent=parent)
#
#         minimum = 2000
#         maximum = 2300
#         interval = 10
#         orientation = Qt.Horizontal
#         labels = None
#
#         levels=range(minimum, maximum+interval, interval)
#         if labels is not None:
#             if not isinstance(labels, (tuple, list)):
#                 raise Exception("<labels> is a list or tuple.")
#             if len(labels) != len(levels):
#                 raise Exception("Size of <labels> doesn't match levels.")
#             self.levels=list(zip(levels,labels))
#         else:
#             self.levels=list(zip(levels,map(str,levels)))
#
#         # if orientation==Qt.Horizontal:
#         #     self.layout=QtWidgets.QVBoxLayout(self)
#         # elif orientation==Qt.Vertical:
#         #     self.layout=QtWidgets.QHBoxLayout(self)
#         # else:
#         #     raise Exception("<orientation> wrong.")
#         #
#         # # gives some space to print labels
#         # self.left_margin=10
#         # self.top_margin=10
#         # self.right_margin=10
#         # self.bottom_margin=10
#         #
#         # self.layout.setContentsMargins(self.left_margin,self.top_margin,
#         #         self.right_margin,self.bottom_margin)
#
#         self.sl=QSlider(orientation, parent=parent)
#         self.sl.setMinimum(minimum)
#         self.sl.setMaximum(maximum)
#         self.sl.setValue(minimum)
#         if orientation==Qt.Horizontal:
#             self.sl.setTickPosition(QtWidgets.QSlider.TicksBelow)
#             self.sl.setMinimumWidth(300) # just to make it easier to read
#         else:
#             self.sl.setTickPosition(QtWidgets.QSlider.TicksLeft)
#             self.sl.setMinimumHeight(300) # just to make it easier to read
#         self.sl.setTickInterval(interval)
#         self.sl.setSingleStep(1)
#
#         x0, y0, wdt, hgt = element["window"].get_position(self.element["position"], self.parent)
#         self.sl.setGeometry(x0, y0, wdt, hgt)
#
#         self.widgets.append(self.sl)
#         # self.layout.addWidget(self.sl)
#
#     def paintEvent(self, e):
#
#         super(LabeledSlider,self).paintEvent(e)
#
#         style=self.sl.style()
#         painter=QPainter(self)
#         st_slider=QStyleOptionSlider()
#         st_slider.initFrom(self.sl)
#         st_slider.orientation=self.sl.orientation()
#
#         length=style.pixelMetric(QStyle.PM_SliderLength, st_slider, self.sl)
#         available=style.pixelMetric(QStyle.PM_SliderSpaceAvailable, st_slider, self.sl)
#
#         for v, v_str in self.levels:
#
#             # get the size of the label
#             rect=painter.drawText(QRect(), Qt.TextDontPrint, v_str)
#
#             if self.sl.orientation()==Qt.Horizontal:
#                 # I assume the offset is half the length of slider, therefore
#                 # + length//2
#                 x_loc=QStyle.sliderPositionFromValue(self.sl.minimum(),
#                         self.sl.maximum(), v, available)+length//2
#
#                 # left bound of the text = center - half of text width + L_margin
#                 left=x_loc-rect.width()//2+self.left_margin
#                 bottom=self.rect().bottom()
#
#                 # enlarge margins if clipping
#                 if v==self.sl.minimum():
#                     if left<=0:
#                         self.left_margin=rect.width()//2-x_loc
#                     if self.bottom_margin<=rect.height():
#                         self.bottom_margin=rect.height()
#
#                     self.layout.setContentsMargins(self.left_margin,
#                             self.top_margin, self.right_margin,
#                             self.bottom_margin)
#
#                 if v==self.sl.maximum() and rect.width()//2>=self.right_margin:
#                     self.right_margin=rect.width()//2
#                     self.layout.setContentsMargins(self.left_margin,
#                             self.top_margin, self.right_margin,
#                             self.bottom_margin)
#
#             else:
#                 y_loc=QStyle.sliderPositionFromValue(self.sl.minimum(),
#                         self.sl.maximum(), v, available, upsideDown=True)
#
#                 bottom=y_loc+length//2+rect.height()//2+self.top_margin-3
#                 # there is a 3 px offset that I can't attribute to any metric
#
#                 left=self.left_margin-rect.width()
#                 if left<=0:
#                     self.left_margin=rect.width()+2
#                     self.layout.setContentsMargins(self.left_margin,
#                             self.top_margin, self.right_margin,
#                             self.bottom_margin)
#
#             pos=QPoint(left, bottom)
#             painter.drawText(pos, v_str)
#
#         return
#
#     def set(self):
#         if self.check_variables():
#             getvar = self.element["getvar"]
#             group = self.element.variable_group
#             name = self.element.variable
#             val = getvar(group, name)
#             self.widgets[0].setValue(val)
#             self.set_dependencies()
#
#     def callback(self):
#         self.okay = True
#         if self.check_variables():
#             newval = self.widgets[0].value()
#
#             # Update value in variable dict
#             if self.okay:
#                 setvar = self.element["setvar"]
#                 group = self.element.variable_group
#                 name = self.element.variable
#                 setvar(group, name, newval)
#                 self.widgets[0].setValue(newval)


from .layer import Layer
from geopandas import GeoDataFrame
from pyogrio import read_dataframe


class CircleLayer(Layer):
    def __init__(self, map, id, map_id, **kwargs):
        super().__init__(map, id, map_id, **kwargs)
        self.color_by_attribute = dict()
        self.legend_items = list()
        pass

    def set_data(
        self, data, color_by_attribute: dict = dict(), legend_items: list = []
    ):
        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                return
            if data.crs != 4326:
                data = data.to_crs(4326)
        else:
            # Read geodataframe from shape file
            data = read_dataframe(data)

        self.color_by_attribute = color_by_attribute
        self.legend_items = legend_items
        self.data = data
        
        self.unit = getattr(self, 'unit', '')
        
        if not self.big_data:
            if len(self.color_by_attribute) == 0:
                # Add new layer
                #TODO: take unit info from settings.toml
                self.map.runjs(
                    "/js/circle_layer.js",
                    "addLayer",
                    arglist=[
                        self.map_id,
                        self.data,
                        self.hover_property,
                        self.min_zoom,
                        self.line_color,
                        self.line_width,
                        self.line_opacity,
                        self.fill_color,
                        self.fill_opacity,
                        self.circle_radius,
                        self.unit
                    ],
                )
            elif len(self.color_by_attribute) > 0:
                # Color by attribute
                self.map.runjs(
                    "/js/circle_layer_custom.js",
                    "addLayer",
                    arglist=[
                        self.map_id,
                        self.data,
                        self.hover_property,
                        self.unit,
                        self.min_zoom,
                        self.color_by_attribute,
                        self.legend_items,
                        self.legend_position,
                        self.legend_title,
                    ],
                )

        self.update()

    def update(self):
        if not self.map.zoom:
            return
        if self.data is None:
            return
        if len(self.data) == 0:
            # Empty GeoDataFrame
            return            
        if self.map.zoom > self.min_zoom and self.big_data and self.visible:
            coords = self.map.map_extent
            xl0 = coords[0][0]
            xl1 = coords[1][0]
            yl0 = coords[0][1]
            yl1 = coords[1][1]
            
            # Limits WGS 84
            gdf = self.data.cx[xl0:xl1, yl0:yl1]
            
            self.unit = getattr(self, 'unit', '')
            if len(self.color_by_attribute) == 0:
                # Add new layer
                self.map.runjs(
                    "/js/circle_layer.js",
                    "addLayer",
                    arglist=[
                        self.map_id,
                        gdf,
                        self.hover_property,
                        self.min_zoom,
                        self.line_color,
                        self.line_width,
                        self.line_opacity,
                        self.fill_color,
                        self.fill_opacity,
                        self.circle_radius,
                        self.unit                        
                    ],
                )
            elif len(self.color_by_attribute) > 0:
                # Color by attribute
                self.map.runjs(
                    "/js/circle_layer_custom.js",
                    "addLayer",
                    arglist=[
                        self.map_id,
                        self.data,
                        self.hover_property,
                        self.unit,
                        self.min_zoom,
                        self.color_by_attribute,
                        self.legend_items,
                        self.legend_position,
                        self.legend_title,
                    ],
                )

    def activate(self):
        self.show()
        self.map.runjs(
            "/js/circle_layer.js",
            "setPaintProperties",
            arglist=[
                self.map_id,
                self.line_color,
                self.line_width,
                self.line_opacity,
                self.fill_color,
                self.fill_opacity,
                self.circle_radius,
            ],
        )
        
        #Not working for color_by_attribute, so fixed with self.update()
        #TODO: make pretty
        self.update()

    def deactivate(self):
        self.map.runjs(
            "/js/circle_layer.js",
            "setPaintProperties",
            arglist=[
                self.map_id,
                self.line_color_inactive,
                self.line_width_inactive,
                self.line_opacity_inactive,
                self.fill_color_inactive,
                self.fill_opacity_inactive,
                self.circle_radius_inactive,
            ],
        )

    def redraw(self):
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data)
        if not self.get_visibility():
            self.set_visibility(False)



from PySide6.QtWidgets import QWidget, QMainWindow, QLabel
from PySide6.QtWidgets import QApplication
from PySide6.QtWidgets import QVBoxLayout
from PySide6 import QtGui
import os

class MainWindow(QMainWindow):
    def __init__(self, window):
        super().__init__()

        self.window = window

        # Determine resize factor for elements (fonts get scaled automatically)
        screen = QApplication.primaryScreen()

        # Create a QLabel widget
        qw = QWidget()
        qw.setVisible(False)
        qw.setGeometry(0, 0, 1, 1)
        qw.show()
        pt = qw.font().pointSize()
        font = QtGui.QFont("times", pt)
        fm = QtGui.QFontMetrics(font)
        height_in_pixels = fm.height()
        resize_factor = height_in_pixels / 11
        self.window.gui.resize_factor = resize_factor

        # Window size
        screen_width = screen.size().width()
        screen_height = screen.size().height()
        self.setWindowTitle(self.window.title)
        self.window_width = min(screen_width - 100, int(self.window.width*self.window.gui.resize_factor))
        self.window_height = min(screen_height - 100, int(self.window.height*self.window.gui.resize_factor))
        min_width = min(screen_width - 100, int(self.window.minimum_width * self.window.gui.resize_factor))
        min_height = min(screen_height - 100, int(self.window.minimum_height * self.window.gui.resize_factor))
        self.setMinimumSize(min_width, min_height)
        if self.window.icon:
          self.setWindowIcon(QtGui.QIcon(self.window.icon))
        # screen = QApplication.primaryScreen()
        self.setGeometry(int(0.5*(screen_width - self.window_width)),
                         int(0.5*(screen_height - self.window_height)),
                         self.window_width,
                         self.window_height)
        if self.window.fixed_size:
            self.setFixedSize(self.window_width, self.window_height)

    def resizeEvent(self, event):
        self.window.resize_elements(self.window.elements)
        if self.window.statusbar:
            self.window.statusbar.set_geometry()

    def closeEvent(self, event):
        QApplication.quit()        




import os
from PyQt5 import QtWebEngineWidgets
from PyQt5 import QtCore, QtWidgets, QtWebChannel
import json
from urllib.request import urlopen
import urllib
from geopandas import GeoDataFrame

from guitares.map.layer import Layer, list_layers, find_layer_by_id


class WebEnginePage(QtWebEngineWidgets.QWebEnginePage):
    def __init__(self, view, print_messages):
        super().__init__(view)
        self.print_messages = print_messages

    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceID):
        if self.print_messages:
            print("javaScriptConsoleMessage: ", level, message, lineNumber, sourceID)


class MapBoxCompare(QtWidgets.QWidget):
    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.gui = element.gui
        self.element = element

        file_name = os.path.join(self.gui.server_path, "js", "mapbox_compare_defaults.js")
        with open(file_name, "w") as f:
            f.write("var default_compare_style = '" + element.map_style + "';\n")
            f.write("var default_compare_center = [" + str(element.map_center[0]) + "," + str(element.map_center[1]) + "]\n")
            f.write("var default_compare_zoom = " + str(element.map_zoom) + ";\n")
            f.write("var default_compare_projection = '" + element.map_projection + "';\n")

        url = "http://localhost:" + str(self.gui.server_port) + "/mapbox_compare.html"
        self.url = url

        self.ready = False
        self.ready_a = False
        self.ready_b = False

        self.server_path = self.gui.server_path

        self.setGeometry(
            0, 0, -1, -1
        )  # this is necessary because otherwise an invisible widget sits over the top left hand side of the screen and block the menu

        view = self.view = QtWebEngineWidgets.QWebEngineView(element.parent.widget)
        channel = self.channel = QtWebChannel.QWebChannel()
        view.page().profile().clearHttpCache()

        self.set_geometry()

        page = WebEnginePage(view, self.gui.js_messages)
        view.setPage(page)
        view.page().setWebChannel(channel)

        channel.registerObject("MapBoxCompare", self)

        view.load(QtCore.QUrl(url))

        self.callback_module = element.module

        self.layer = {}
        self.map_extent = None
        self.map_moved = None
        self.point_clicked_callback = None
        self.zoom = None

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.reload)
        self.timer.setSingleShot(True)
        self.timer.start(1000)

    def reload(self):
        print("Reloading ...")
        self.view.page().setWebChannel(self.channel)
        self.channel.registerObject("MapBoxCompare", self)
        self.view.load(QtCore.QUrl(self.url))

    def set(self):
        pass

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.view.setGeometry(x0, y0, wdt, hgt)

    def take_screenshot(self, output_file):
        self.view.grab().save(output_file, b"PNG")

    @QtCore.pyqtSlot(str)
    def mapReady(self, inpstr):
        inp = json.loads(inpstr)
        coords = inp[0:2]
        self.map_extent = coords
        side = inp[2]
        if side == "a":
            self.ready_a = True
        elif side == "b":
            self.ready_b = True
        if self.ready_a and self.ready_b:
            self.ready = True
        if self.ready:
            if hasattr(self.callback_module, "map_ready"):
                self.callback_module.map_ready()

    @QtCore.pyqtSlot(str)
    def layerStyleSet(self, coords):
        self.redraw_layers()

    # @QtCore.pyqtSlot(str)
    # def mouseMoved(self, coords):
    #     coords = json.loads(coords)
    #     self.map_extent = coords
    #     # Loop through layers to update each
    #     layers = list_layers(self.layer)
    #     for layer in layers:
    #         layer.update()
    #     if hasattr(self.callback_module, "map_moved"):
    #         self.callback_module.map_moved(coords)

    @QtCore.pyqtSlot(str)
    def mapMoved(self, coords):
        coords = json.loads(coords)
        self.map_extent = coords[0:2]
        self.map_center = coords[2:5]
        self.zoom = coords[4]
        # Loop through layers to update each
        layers = list_layers(self.layer)
        for layer in layers:
            layer.update()
        if hasattr(self.callback_module, "map_moved"):
            self.callback_module.map_moved(coords)

    # @QtCore.pyqtSlot(str)
    # def pointClicked(self, coords):
    #     coords = json.loads(coords)
    #     if self.point_clicked_callback:
    #         self.point_clicked_callback(coords)

    # @QtCore.pyqtSlot(str)
    # def getMapExtent(self, coords):
    #     coords = json.loads(coords)
    #     self.map_extent = coords

    # @QtCore.pyqtSlot(str, str)
    # def featureClicked(self, layer_id, feature_props):
    #     # Find layer by ID
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     if hasattr(layer, "select"):
    #         if layer.select:
    #             layer.select(json.loads(feature_props))

    # @QtCore.pyqtSlot(str, str, str)
    # def featureDrawn(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_drawn(json.loads(feature_collection), feature_id)

    # @QtCore.pyqtSlot(str, str, str)
    # def featureModified(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_modified(json.loads(feature_collection), feature_id)

    # @QtCore.pyqtSlot(str, str, str)
    # def featureSelected(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_selected(json.loads(feature_collection), feature_id)

    # @QtCore.pyqtSlot(str)
    # def featureDeselected(self, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     if layer:
    #         layer.feature_deselected()

    # @QtCore.pyqtSlot(str, str, str)
    # def featureAdded(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_added(json.loads(feature_collection), feature_id)

    # def get_extent(self):
    #     js_string = "import('/js/main.js').then(module => {module.getExtent()});"
    #     self.view.page().runJavaScript(js_string)

    # def click_point(self, callback):
    #     self.point_clicked_callback = callback
    #     self.runjs("/js/main.js", "clickPoint")

    # def set_center(self, lon, lat):
    #     self.runjs("/js/main.js", "setCenter", arglist=[lon, lat])

    # def set_zoom(self, zoom):
    #     self.runjs("/js/main.js", "setZoom", arglist=[zoom])

    # def fit_bounds(self, lon1, lat1, lon2, lat2):
    #     self.runjs("/js/main.js", "fitBounds", arglist=[lon1, lat1, lon2, lat2])

    def jump_to(self, lon, lat, zoom):
        self.runjs("/js/compare.js", "jumpTo", arglist=[lon, lat, zoom])

    def fly_to(self, lon, lat, zoom):
        self.runjs("/js/compare.js", "flyTo", arglist=[lon, lat, zoom])

    # def set_projection(self, projection):
    #     self.runjs("/js/main.js", "setProjection", arglist=[projection])

    def set_layer_style(self, style):
        self.runjs("/js/compare.js", "setLayerStyle", arglist=[style])

    def set_slider(self, npix):
        self.runjs("/js/compare.js", "setSlider", arglist=[npix])

    # def set_terrain(self, true_or_false, exaggeration):
    #     self.runjs("/js/main.js", "setTerrain", arglist=[true_or_false, exaggeration])

    # def set_mouse_default(self):
    #     self.runjs("/js/draw.js", "setMouseDefault", arglist=[])

    def add_layer(self, layer_id):
        # Adds a container layer
        if layer_id not in self.layer:
            self.layer[layer_id] = Layer(self, layer_id, layer_id)
            self.layer[layer_id].map_id = layer_id
        else:
            print("Layer " + layer_id + " already exists.")
        return self.layer[layer_id]

    def list_layers(self):
        # Return a list with all layers
        return list_layers(self.layer)

    def redraw_layers(self):
        # Redraw all layers (after map style has changed)
        layers = self.list_layers()
        for layer in layers:
            layer.redraw()

    def runjs(self, module, function, arglist=None):
        if not arglist:
            arglist = []
        string = "import('" + module + "').then(module => {module." + function + "("
        for iarg, arg in enumerate(arglist):
            if isinstance(arg, bool):
                if arg:
                    string = string + "true"
                else:
                    string = string + "false"
            elif isinstance(arg, int):
                string = string + str(arg)
            elif isinstance(arg, float):
                string = string + str(arg)
            elif isinstance(arg, dict):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, list):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, GeoDataFrame):
                if len(arg) == 0:
                    string = string + "{}"
                else:
                    string = string + arg.to_json()
            else:
                string = string + "'" + arg + "'"
            if iarg < len(arglist) - 1:
                string = string + ","
        string = string + ")});"
        self.view.page().runJavaScript(string)



import os
import glob
from PIL import Image
import matplotlib
from matplotlib import cm
import matplotlib.pyplot as plt
import rasterio
# import rasterio.features
from rasterio.warp import calculate_default_transform, reproject, Resampling, transform_bounds
from rasterio import MemoryFile
# from rasterio.transform import Affine
from matplotlib.colors import LightSource
import numpy as np
import copy
import xarray as xr
from pyproj import Transformer

from .colorbar import ColorBar
from guitares.colormap import cm2png

from .layer import Layer

class RasterLayer(Layer):
    def __init__(self, map, id, map_id):
        super().__init__(map, id, map_id)
        self.active = False
        self.type   = "raster"
        self.new    = True
        self.file_name = map_id + ".png"

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False


    def clear(self):
        self.active = False
        self.map.runjs("/js/main.js", "removeLayer", arglist=[self.map_id])

    def update(self):
        print("Updating image layer")

    def set_data(self,
                 x=None,
                 y=None,
                 z=None,
                 image_file=None,
                 legend_title="",
                 cmin=None,
                 cmax=None,
                 cstep=None,
                 decimals=None,
                 crs=None,
                 colormap=None):
        
        self.data = {}
        self.data["x"] = x
        self.data["y"] = y
        self.data["z"] = z
        self.data["image_file"] = image_file
        self.data["legend_title"] = legend_title
        self.data["cmin"] = cmin
        self.data["cmax"] = cmax
        self.data["cstep"] = cstep
        self.data["decimals"] = decimals
        self.data["crs"] = crs
        self.data["colormap"] = colormap

        cmap = colormap

        if not crs:
            src_crs = "EPSG:4326"
        else:
            src_crs = "EPSG:" + str(crs.to_epsg())

        # Web Mercator
        dst_crs = 'EPSG:3857'

        if image_file:

            with rasterio.open(image_file) as src:
                transform, width, height = calculate_default_transform(
                    src.crs, dst_crs, src.width, src.height, *src.bounds)
                kwargs = src.meta.copy()
                kwargs.update({
                    'crs': dst_crs,
                    'transform': transform,
                    'width': width,
                    'height': height
                })
                # bnds = src.bounds

                mem_file = MemoryFile()
                with mem_file.open(**kwargs) as dst:
                    for i in range(1, src.count + 1):
                        reproject(
                            source=rasterio.band(src, i),
                            destination=rasterio.band(dst, i),
                            src_transform=src.transform,
                            src_crs=src.crs,
                            dst_transform=transform,
                            dst_crs=dst_crs,
                            resampling=Resampling.nearest)

                    band1 = dst.read(1)

            new_bounds = transform_bounds(dst_crs, src_crs,
                                          dst.bounds[0],
                                          dst.bounds[1],
                                          dst.bounds[2],
                                          dst.bounds[3])
            isn = np.where(band1 < 0.001)
            band1[isn] = np.nan

            band1 = np.flipud(band1)
            cminimum = np.nanmin(band1)
            cmaximum = np.nanmax(band1)

            norm = matplotlib.colors.Normalize(vmin=cminimum, vmax=cmaximum)
            vnorm = norm(band1)

            cmap = cm.get_cmap(colormap)
            im = Image.fromarray(np.uint8(cmap(vnorm) * 255))

            overlay_file = "./overlays/" + self.file_name
            im.save(os.path.join(self.map.server_path, overlay_file))

            # Bounds
            bounds = [[new_bounds[0], new_bounds[2]], [new_bounds[3], new_bounds[1]]]

        else:

            # Color scale
            if self.color_scale_auto:
                if self.color_scale_symmetric:
                    if self.color_scale_symmetric_side == "min":
                        cmin = np.nanmin(z)
                        if cmin > 0:
                            cmin = -cmin
                        cmax = -cmin
                    elif self.color_scale_symmetric_side == "max":
                        cmax = np.nanmax(z)
                        if cmax < 0:
                            cmax = -cmax
                        cmin = -cmax
                    else:
                        cmx = max(abs(np.nanmin(z)), abs(np.nanmax(z)))
                        cmin = -cmx
                        cmax = cmx
                else:
                    cmin = np.nanmin(z)
                    cmax = np.nanmax(z)
                if cmax < cmin + 0.01:
                    cmin = -0.01
                    cmax = 0.01    
            else:    
                cmin = self.color_scale_cmin
                cmax = self.color_scale_cmax

            if self.hillshading:
                ls = LightSource(azdeg=315, altdeg=30)
                dx = (x[1] - x[0]) / 2
                dy = (y[1] - y[0]) / 2
                rgb = ls.shade(np.flipud(z), cmap,
                            vmin=cmin,
                            vmax=cmax,
                            dx=dx * 0.5,
                            dy=dy * 0.5,
                            vert_exag=10.0,
                            blend_mode="soft")
                rgb = rgb * 255

            else:
                norm = matplotlib.colors.Normalize(vmin=cmin, vmax=cmax)
                vnorm = norm(np.flipud(z))
                cmap = cm.get_cmap(colormap)
                rgb = cmap(vnorm) * 255

            # Create xarray DataArray with RGBA values of the rgb array
            rgba_da = xr.DataArray(np.transpose(rgb, (2, 0, 1)),
                                dims=["band", "y", "x"],
                                coords={ "band": [0, 1, 2, 3], "y": np.flip(y), "x": x})

            # 1. Assign CRS and spatial dims
            rgba_da.rio.set_spatial_dims(x_dim="x", y_dim="y", inplace=True)
            rgba_da.rio.write_crs(src_crs, inplace=True)  # e.g., EPSG:32633

            # 2. Reproject to Web Mercator
            rgba_3857 = rgba_da.rio.reproject("EPSG:3857")

            # 3. Export as PNG
            # Convert to uint8 and rearrange shape for Pillow
            # There may be NaN values in the data that cannot be converted to uint8
            # So we need to fill them with 0 (transparent)
            # and convert to uint8
            rgba_3857 = rgba_3857.fillna(0)
            rgba_3857 = rgba_3857.clip(min=0, max=255)
            # Convert to uint8
            img_data = rgba_3857.values.astype(np.uint8)  # shape: (height, width, 4)

            # In case of xarray with shape (band, y, x)
            if img_data.shape[0] == 4:
                img_data = np.transpose(img_data, (1, 2, 0))  # to (y, x, 4)

            image = Image.fromarray(img_data, mode="RGBA")

            # overlay_file = os.path.join(self.map.server_path, "testing.png")
            image.save(os.path.join(self.map.server_path, "overlays", self.file_name))

            # Get bounds in EPSG:3857
            left, bottom, right, top = rgba_3857.rio.bounds()

            # Reproject bounds to EPSG:4326
            transformer = Transformer.from_crs("EPSG:3857", "EPSG:4326", always_xy=True)
            west, south = transformer.transform(left, bottom)
            east, north = transformer.transform(right, top)

            overlay_file = "./overlays/" + self.file_name

            # It's possible that the dst.bounds cover the entire globe, and that bounds will
            # then run from 179.888 to 179.999
            # In that case bounds[0][0] = bounds[0][0] - 360.0
            if left < -19800000.0 and right > 19800000.0:
                if west > 0.0:
                    west = -180.0
                if east < 0.0:
                    east = 180.0    

            if west > east:
                west -= 360.0
            if west < -180.0:
                west += 360.0
                east += 360.0

            # Delete old legend files
            for file_name in glob.glob(os.path.join(self.map.server_path, "overlays", self.map_id + ".legend.*.png")):
                try:
                    os.remove(file_name)
                except:
                    pass

            # add random integer string to legend file to force reload                
            # create string with random integer between 1 and 1,000,000
            rstring = str(np.random.randint(1, 1000000))
            legend_file = self.map_id + ".legend." + rstring + ".png"
            width = 1.0
            height = 2.0
            cm2png(cmap,
                file_name = os.path.join(self.map.server_path, "overlays", legend_file),
                orientation="vertical",
                vmin=cmin,
                vmax=cmax,
                width=width,
                height=height)

            # Legend
            clrbar = ColorBar(colormap=colormap, legend_title=legend_title)
            clrbar.make(cmin, cmax, cstep=cstep, decimals=decimals)
            clrmap_string = clrbar.to_json()

            clrmap_string = "./overlays/" + legend_file

        # print(bounds)

        if east > 180.0:
            # Image crosses date line
            # Need to create two images, west and east
            # First read in the png file
            im = Image.open(os.path.join(self.map.server_path, overlay_file))
            # find pixel index
            npixx = im.size[0]
            dlon = east - west
            ipixx = int(npixx * (180.0 - west) / dlon)
            im1 = im.crop((0, 0, ipixx, im.size[1]))
            im2 = im.crop((ipixx + 1, 0, im.size[0], im.size[1]))
            im1.save(os.path.join(self.map.server_path, overlay_file.replace(".png", ".a.png")))
            im2.save(os.path.join(self.map.server_path, overlay_file.replace(".png", ".b.png")))
            overlay_file_a = overlay_file.replace(".png", ".a.png")
            overlay_file_b = overlay_file.replace(".png", ".b.png")

            west_a = west
            east_a = 180.0
            west_b = -180.0
            east_b = east - 360.0

            bounds_a = [[west_a, east_a], [south, north]]
            bounds_b = [[west_b, east_b], [south, north]]

            split_image = True
        else:
            bounds = [[west, east], [south, north]]
            split_image = False

        if self.new:

            # First time we add the layer, so we need to create it. Always create both west and east layers.

            # Add west layer
            self.map.runjs("/js/image_layer.js", "addLayer", arglist=[self.map_id + ".a", self.side])
            # Add east layer
            self.map.runjs("/js/image_layer.js", "addLayer", arglist=[self.map_id + ".b", self.side])

        if split_image:

            # Make the east layer visible
            self.map.runjs(self.main_js, "showLayer", arglist=[self.map_id + ".b", self.side])

            # Now we need to update the layers with the new bounds and opacity
            self.map.runjs("/js/image_layer.js", "updateLayer", arglist=[overlay_file_a,
                                                                         self.map_id + ".a",                                                                        
                                                                         bounds_a,
                                                                         clrmap_string])
            self.map.runjs("/js/image_layer.js",
                           "setOpacity",
                           arglist=[self.map_id + ".a", self.opacity, self.side])

            self.map.runjs("/js/image_layer.js", "updateLayer", arglist=[overlay_file_b,
                                                                         self.map_id + ".b",                                                                        
                                                                         bounds_b,
                                                                         ""])
            self.map.runjs("/js/image_layer.js",
                           "setOpacity",
                           arglist=[self.map_id + ".b", self.opacity, self.side])

        else:
            # Just update the layer with the new bounds and opacity
            self.map.runjs("/js/image_layer.js", "updateLayer", arglist=[overlay_file,
                                                                         self.map_id + ".a",                                                                         
                                                                         bounds,
                                                                         clrmap_string])
            self.map.runjs("/js/image_layer.js",
                           "setOpacity",
                           arglist=[self.map_id + ".a", self.opacity, self.side])

            # Make the east layer invisible
            self.map.runjs(self.main_js, "hideLayer", arglist=[self.map_id + ".b", self.side])

        self.new = False

    def redraw(self):
        if self.data:
            self.new = True
            self.set_data(x=self.data["x"],
                          y=self.data["y"],
                          z=self.data["z"],
                          image_file=self.data["image_file"],
                          legend_title=self.data["legend_title"],
                          cmin=self.data["cmin"],
                          cmax=self.data["cmax"],
                          cstep=self.data["cstep"],
                          decimals=self.data["decimals"],
                          crs=self.data["crs"],
                          colormap=self.data["colormap"])
        if not self.get_visibility():
            self.set_visibility(False)
        



from PyQt5 import QtWebEngineWidgets
from PyQt5 import QtCore, QtWidgets

#from guitares.gui import get_position

class WebEnginePage(QtWebEngineWidgets.QWebEnginePage):
    def __init__(self, view, print_messages):
        super().__init__(view)
        self.print_messages = print_messages
    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceID):
        if self.print_messages:
            print("javaScriptConsoleMessage: ", level, message, lineNumber, sourceID)


class WebPage(QtWidgets.QWidget):
    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.element = element

        view = self.view = QtWebEngineWidgets.QWebEngineView(element.parent.widget)

        self.set_geometry()

        page = WebEnginePage(view, element.gui.js_messages)
        view.setPage(page)

        if type(self.element.url) == str:
            url = self.element.url
        else:
            url = self.element.getvar(element.url.variable_group, element.url.variable)    
        url = url.replace('\\', '/')
        view.load(QtCore.QUrl(url))

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.reload)
        self.timer.setSingleShot(True)
        self.timer.start(500)

    def set(self):
        pass

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.view.setGeometry(x0, y0, wdt, hgt)

    def reload(self):
        if type(self.element.url) == str:
            url = self.element.url
        else:
            url = self.element.getvar(self.element.url.variable_group, self.element.url.variable)    
        url = url.replace('\\', '/')
        self.view.load(QtCore.QUrl(url))

    def set_url(self, url):
        self.element.url = url.replace('\\', '/')
        self.view.load(QtCore.QUrl(self.element.url))

    def take_screenshot(self, output_file):
        self.view.grab().save(output_file, b'PNG')
